@@ -2,23 +2,25 @@
 #
 # SPDX-License-Identifier: Apache-2.0
 
-"""Semantic Markdown Splitter
-"""
+"""Semantic Markdown Splitter."""
 
-from logging import getLogger
 import re
-from typing import Tuple, Type, TypedDict, Optional
-import tiktoken
+from logging import getLogger
+from typing import TYPE_CHECKING, Optional, TypedDict
+
 import mdformat
-import spacy
-from mistletoe.token import Token
-from mistletoe import block_token, markdown_renderer, span_token
+import tiktoken
 from mistletoe import Document as MisDocument
-from mistletoe.markdown_renderer import MarkdownRenderer
+from mistletoe import block_token, markdown_renderer, span_token
+from mistletoe.token import Token
 
 from wurzel.datacontract import MarkdownDataContract
 from wurzel.exceptions import MarkdownException
+from wurzel.utils.markdown_table_splitter import MarkdownTableSplitterUtil
+from wurzel.utils.to_markdown.html2md import MD_RENDER_LOCK
 
+if TYPE_CHECKING:
+    import spacy
 LEVEL_MAPPING = {
     block_token.Heading: 0,  # actually 1-6
     block_token.List: 7,
@@ -31,20 +33,19 @@ LEVEL_MAPPING = {
     block_token.ThematicBreak: 14,
 }
 LEVEL_UNDEFINED = 15
-OPENAI_ENCODING = tiktoken.encoding_for_model("gpt-3.5-turbo")
 
 log = getLogger(__name__)
 
 
 class MetaDataDict(TypedDict):
-    """Dict definition of metadata"""
+    """Dict definition of metadata."""
 
     keywords: str
     url: str
 
 
 class DocumentNode(TypedDict):
-    """Dict definition of internal tree structure for splitting"""
+    """Dict definition of internal tree structure for splitting."""
 
     highest_level: int
     token_len: int
@@ -53,18 +54,6 @@ class DocumentNode(TypedDict):
     metadata: MetaDataDict
 
 
-def _get_token_len(text: str) -> int:
-    """Get OpenAI Token length
-
-    Args:
-        text (str): Test to encode
-
-    Returns:
-        int: count of tokens
-    """
-    return len(OPENAI_ENCODING.encode(text))
-
-
 def _is_all_children_same_level(children: list[DocumentNode]) -> bool:
     nbr_different_level = len({c["highest_level"] for c in children})
     return nbr_different_level == 1
@@ -91,12 +80,12 @@ def _get_highest_level_of_children(children: list[DocumentNode]) -> int:
 
 
 def _get_heading_text(token: block_token.Heading):
-    """Get heading text from Mistletoe heading block token"""
+    """Get heading text from Mistletoe heading block token."""
     if token.content:
         return token.content
     if "children" in vars(token) and len(token.children) > 0:
         return " ".join([c.content for c in token.children if hasattr(c, "content")])
-    return ''
+    return ""
 
 
 def _is_standalone_a_heading(text):
@@ -106,20 +95,10 @@ def _is_standalone_a_heading(text):
     return isinstance(childs[0], block_token.Heading)
 
 
-def _cut_to_tokenlen(text: str, token_len: int) -> str:
-    """Cut Text to token length using OpenAI"""
-
-    tokens_old = OPENAI_ENCODING.encode(text)
-    if len(tokens_old) > token_len:
-        tokens = tokens_old[:token_len]
-        return OPENAI_ENCODING.decode(tokens)
-    return OPENAI_ENCODING.decode(tokens_old)
-
-
 def _format_markdown_docs(
     docs: list[MarkdownDataContract],
 ) -> list[MarkdownDataContract]:
-    """Formats the Markdown Document by the standards"""
+    """Formats the Markdown Document by the standards."""
     return [
         MarkdownDataContract(
             md=mdformat.text(doc.md).strip(),
@@ -131,65 +110,142 @@ def _format_markdown_docs(
 
 
 class WurzelMarkdownRenderer(markdown_renderer.MarkdownRenderer):
-    """Fix For markdown_renderer.MarkdownRenderer"""
-    #pylint: disable=unused-argument, arguments-differ
-    def render_table_cell(self, token: block_token.TableCell, max_line_length:int) -> str:
+    """Fix For markdown_renderer.MarkdownRenderer."""
+
+    # pylint: disable=unused-argument, arguments-differ
+    def render_table_cell(self, token: block_token.TableCell, max_line_length: int) -> str:
+        """Renders the content of a table cell.
+
+        Args:
+            token (block_token.TableCell): The table cell token to render.
+            max_line_length (int): The maximum allowed line length for the content.
+
+        Returns:
+            str: The rendered content of the table cell.
+
+        """
         return self.render_inner(token)
 
-    def render_table_row(self, token: block_token.TableRow, max_line_length:int) -> str:
+    def render_table_row(self, token: block_token.TableRow, max_line_length: int) -> str:
+        """Renders a table row from the given token.
+
+        Args:
+            token (block_token.TableRow): The table row token to be rendered.
+            max_line_length (int): The maximum allowed line length for the rendered row.
+
+        Returns:
+            str: The rendered table row as a string.
+
+        """
         return self.render_inner(token)
+
+
 class SemanticSplitter:
-    """Splitter implementation"""
+    """Splitter implementation."""
 
-    nlp: spacy.language.Language
+    nlp: "spacy.language.Language"
     token_limit: int
     token_limit_buffer: int
     token_limit_min: int
+    tokenizer_model_encoding: tiktoken.Encoding
+    repeat_table_header_row: bool
 
+    # pylint: disable=too-many-positional-arguments
     def __init__(
         self,
         token_limit: int = 256,
         token_limit_buffer: int = 32,
         token_limit_min: int = 64,
-        spacy_model:str = "de_core_news_sm"
-    ) -> None:
+        spacy_model: str = "de_core_news_sm",
+        repeat_table_header_row: bool = True,
+        tokenizer_model: str = "gpt-3.5-turbo",
+    ) -> None:  # pylint: enable=too-many-positional-arguments
+        """Initializes the SemanticSplitter class with specified token limits and a spaCy language model.
+
+        Args:
+            token_limit (int, optional): The maximum number of tokens allowed. Defaults to 256.
+            token_limit_buffer (int, optional): The buffer size for token limit to allow flexibility. Defaults to 32.
+            token_limit_min (int, optional): The minimum number of tokens required. Defaults to 64.
+            spacy_model (str, optional): The name of the spaCy language model to load. Defaults to "de_core_news_sm".
+            repeat_table_header_row (bool, optional): If a table is splitted, the header is repeated. Defaults to True.
+            tokenizer_model (str, optional): The name of the tokenizer model to use for encoding. Defaults to "gpt-3.5-turbo".
+
+        Raises:
+            OSError: If the specified spaCy model cannot be loaded.
+
+        """
+        import spacy  # pylint: disable=import-outside-toplevel
+
         self.nlp = spacy.load(spacy_model)
         self.token_limit = token_limit
         self.token_limit_buffer = token_limit_buffer
         self.token_limit_min = token_limit_min
+        self.repeat_table_header_row = repeat_table_header_row
+        self.tokenizer_model_encoding = tiktoken.encoding_for_model(tokenizer_model)
+
+    def _get_token_len(self, text: str) -> int:
+        """Get Token length.
+
+        Args:
+            text (str): Text to encode
+
+        Returns:
+            int: count of tokens
+
+        """
+        return len(self.tokenizer_model_encoding.encode(text))
+
+    def _cut_to_tokenlen(self, text: str, token_len: int, return_discarded_text: bool = False) -> str | tuple[str, str]:
+        """Cut text to max. token length using OpenAI tokenizer.
+
+        Args:
+            text (str): Input text
+            token_len (int): Max. number of tokens for output text
+            return_discarded_text (bool): If enabled, the function returns also the discarded text (beyond token limit). Defaults to False.
+
+        Returns:
+            str | tuple[str, str]: Text limited to max. token count (and the discarded text, depending on `return_discarded_text`)
+
+        """
+        input_tokens = self.tokenizer_model_encoding.encode(text)
+
+        output_tokens = input_tokens[:token_len]  # ensure token length
+        output_text = self.tokenizer_model_encoding.decode(output_tokens)  # convert back to text
+
+        if return_discarded_text:
+            discarded_tokens = input_tokens[token_len:]  # beyond token length
+            discarded_text = self.tokenizer_model_encoding.decode(discarded_tokens)  # convert back to text
+
+            return output_text, discarded_text
+
+        return output_text
 
     def _is_short(self, text: str) -> bool:
-        return _get_token_len(text) <= self.token_limit - self.token_limit_buffer
+        return self._get_token_len(text) <= self.token_limit - self.token_limit_buffer
 
-    def _is_table(self,doc:DocumentNode)->bool:
+    def _is_table(self, doc: DocumentNode) -> bool:
         return doc["highest_level"] == LEVEL_MAPPING[block_token.Table]
 
-
     def _is_within_targetlen_w_buffer(self, text: str) -> bool:
-        length = _get_token_len(text)
-        return (
-            self.token_limit + self.token_limit_buffer
-            >= length
-            >= self.token_limit - self.token_limit_buffer
-        )
+        length = self._get_token_len(text)
+        return self.token_limit + self.token_limit_buffer >= length >= self.token_limit - self.token_limit_buffer
 
     def _render_doc(self, doc: MisDocument) -> str:
-        """Render Mistletoe Markdown Document"""
+        """Render Mistletoe Markdown Document."""
         try:
-            with WurzelMarkdownRenderer() as renderer:
+            with MD_RENDER_LOCK, WurzelMarkdownRenderer() as renderer:
                 return renderer.render(doc)  # type: ignore[no-any-return]
         except Exception as e:
             raise MarkdownException(e) from e
 
     def _get_custom_level(self, block: block_token.BlockToken) -> int:
-        """Get the hierarchical level for a mistletoe node"""
-
+        """Get the hierarchical level for a mistletoe node."""
         if isinstance(block, block_token.Heading):
             return int(block.level)
         return LEVEL_MAPPING.get(type(block), LEVEL_UNDEFINED)
 
     def _merge_children(self, children: list[DocumentNode]) -> MisDocument:
-        """Create a document out of a list of children"""
+        """Create a document out of a list of children."""
         new_doc = MisDocument([])
         # If all children a span tokens add them to a paragraph
         # because problems otherwise
@@ -201,25 +257,21 @@ class SemanticSplitter:
             new_doc.children += children
         return new_doc
 
-    def _find_highest_level(
-        self, children: list[DocumentNode], min_level: int = 0
-    ) -> Tuple[int, Optional[Token], Optional[DocumentNode]]:
-        """
-        Among a list of children nodes find the one with the highest level.
-        Return a tuple of that level, level node type and that child
+    def _find_highest_level(self, children: list[DocumentNode], min_level: int = 0) -> tuple[int, Optional[Token], Optional[DocumentNode]]:
+        """Among a list of children nodes find the one with the highest level.
+
+        Return a tuple of that level, level node type and that child.
         """
 
-        def is_any_children(
-            children: list[DocumentNode], block_type: Type[block_token.BlockToken]
-        ):
-            """Check if any Mistletoe Node (child) is of specific type"""
+        def is_any_children(children: list[DocumentNode], block_type: type[block_token.BlockToken]):
+            """Check if any Mistletoe Node (child) is of specific type."""
             for child in children:
                 if isinstance(child, block_type):
                     return True
             return False
 
         highest_level: int = LEVEL_UNDEFINED
-        highest_type: Optional[Type[Token]] = None
+        highest_type: Optional[type[Token]] = None
         highest_element: Optional[DocumentNode] = None
 
         if children is None:
@@ -244,27 +296,16 @@ class SemanticSplitter:
             span_token.LineBreak,  # Add table
             block_token.ThematicBreak,
         ]:
-            if (
-                (highest_level == LEVEL_UNDEFINED)
-                and is_any_children(children, block_type)
-                and (LEVEL_MAPPING[block_type] > min_level)
-            ):
+            if (highest_level == LEVEL_UNDEFINED) and is_any_children(children, block_type) and (LEVEL_MAPPING[block_type] > min_level):
                 highest_level = LEVEL_MAPPING[block_type]
                 highest_type = block_type
-                highest_element = [
-                    child for child in children if isinstance(child, block_type)
-                ][0]
+                highest_element = [child for child in children if isinstance(child, block_type)][0]
                 break
 
         return highest_level, highest_type, highest_element
 
-    def _split_children(
-        self, children: list[MisDocument], min_level: int = 0
-    ) -> list[MisDocument]:
-        """
-        Split a list of children in the most semantic way
-        and return a list of Documents with them merged
-        """
+    def _split_children(self, children: list[MisDocument], min_level: int = 0) -> list[MisDocument]:
+        """Split a list of children in the most semantic way and return a list of Documents with them merged."""
         if len(children) == 1:
             if "children" in vars(children[0]) or "_children" in vars(children[0]):
                 return self._split_children(children[0].children)
@@ -307,14 +348,12 @@ class SemanticSplitter:
         """Split a text into sentences using a NLP model.
 
         This does not use a Regex based approach on purpose as they break with
-        punctuation very easily see: https://stackoverflow.com/a/61254146"""
+        punctuation very easily see: https://stackoverflow.com/a/61254146
+        """
         return [sentence_span.text for sentence_span in self.nlp(text).sents]
 
-    def _markdown_hierarchy_parser(
-        self, text: str, metadata: MetaDataDict, max_depth: int = 30
-    ) -> DocumentNode:
-        """
-        Splits a Markdown string into a semantic Markdown based hierarchy
+    def _markdown_hierarchy_parser(self, text: str, metadata: MetaDataDict, max_depth: int = 30) -> DocumentNode:
+        """Splits a Markdown string into a semantic Markdown based hierarchy.
 
         Given a Markdown string it hierarchically splits that text using
         the semantic information from the Markdown document until
@@ -325,7 +364,7 @@ class SemanticSplitter:
         """
         md_doc = MisDocument(text)
         highest_level, _, _ = self._find_highest_level(md_doc.children)
-        token_len = _get_token_len(text)
+        token_len = self._get_token_len(text)
 
         # Reached max recursion depth
         if max_depth == 0:
@@ -342,10 +381,7 @@ class SemanticSplitter:
         splits: MisDocument = self._split_children(md_doc.children)
 
         def has_node_a_known_level(x):
-            return any(
-                isinstance(x.children[0] if isinstance(x, MisDocument) else x, cl)
-                for cl in LEVEL_MAPPING
-            )
+            return any(isinstance(x.children[0] if isinstance(x, MisDocument) else x, cl) for cl in LEVEL_MAPPING)
 
         no_child_has_a_known_level = not any(has_node_a_known_level(x) for x in splits)
         # No child has a known level this means we only have sentences and no more semantic information
@@ -358,7 +394,7 @@ class SemanticSplitter:
                 children=[
                     DocumentNode(
                         highest_level=highest_level,
-                        token_len=_get_token_len(sent),
+                        token_len=self._get_token_len(sent),
                         text=sent,
                         metadata=metadata,
                         children=[],
@@ -378,12 +414,8 @@ class SemanticSplitter:
                 DocumentNode(
                     highest_level=highest_level_child,
                     text=md_child,
-                    token_len=_get_token_len(md_child),
-                    children=[
-                        self._markdown_hierarchy_parser(
-                            md_child, metadata, max_depth - 1
-                        )
-                    ],
+                    token_len=self._get_token_len(md_child),
+                    children=[self._markdown_hierarchy_parser(md_child, metadata, max_depth - 1)],
                     metadata=metadata,
                 )
             )
@@ -397,19 +429,17 @@ class SemanticSplitter:
 
     # unused
     def _split_by_sentence(self, text: str) -> list[str]:
-        """sometimes _split_children does not find children leafs with are smaller then TOKEN_LIMIT.
-        Thus we need to split by sentence"""
+        """Sometimes _split_children does not find children leafs with are smaller then TOKEN_LIMIT.
+
+        Thus we need to split by sentence.
+        """
         token_limit = self.token_limit
         token_buffer = self.token_limit_buffer
-        lenth = _get_token_len(text)
+        lenth = self._get_token_len(text)
         needed_splits = lenth // token_limit
         if not needed_splits:
             return [text]
-        sentences = [
-            (_get_token_len(sent), f"{sent}. ")
-            for sent in re.split(r"\.(?=\s|\\n)", text)
-            if sent.strip()
-        ]
+        sentences = [(self._get_token_len(sent), f"{sent}. ") for sent in re.split(r"\.(?=\s|\\n)", text) if sent.strip()]
         chunks: list[str] = []
         chunk = ""
         chunk_len = 0
@@ -419,7 +449,7 @@ class SemanticSplitter:
                     chunks.append(chunk)
                     chunk = ""
                     chunk_len = 0
-                chunks.append(_cut_to_tokenlen(sent, token_limit))  # cut this
+                chunks.append(self._cut_to_tokenlen(sent, token_limit))  # cut this
                 # Last piece of sentence is discarded
                 continue
             if chunk_len + size > token_limit + token_buffer:  # with next to big
@@ -428,31 +458,25 @@ class SemanticSplitter:
                 chunk_len = 0
             chunk += sent
             chunk_len += size
-            if (
-                token_limit + token_buffer >= chunk_len >= token_limit - token_buffer
-            ):  # together they fit
+            if token_limit + token_buffer >= chunk_len >= token_limit - token_buffer:  # together they fit
                 chunks.append(chunk)
 
                 chunk = ""
                 chunk_len = 0
         if chunk:
-            chunks.append(_cut_to_tokenlen(chunk, token_limit))
+            chunks.append(self._cut_to_tokenlen(chunk, token_limit))
         # chunks = [
         #    (chunk.replace("\n").strip() if not "#" else chunk.strip()) for chunk in chunks
         # ] This was broken
         chunks = [
-            (
-                _cut_to_tokenlen(chunk, token_limit)
-                if not self._is_within_targetlen_w_buffer(chunk)
-                or self._is_short(chunk)
-                else chunk
-            )
+            (self._cut_to_tokenlen(chunk, token_limit) if not self._is_within_targetlen_w_buffer(chunk) or self._is_short(chunk) else chunk)
             for chunk in chunks
         ]
         for chunk in chunks:
             assert self._is_within_targetlen_w_buffer(chunk) or self._is_short(chunk)
         return chunks
-    #pylint: disable-next=too-many-positional-arguments
+
+    # pylint: disable-next=too-many-positional-arguments
     def _handle_parsing_of_children(
         self,
         doc: DocumentNode,
@@ -461,18 +485,23 @@ class SemanticSplitter:
         remaining_snipped: str,
         recursive_depth: int,
     ) -> tuple[str, list[MarkdownDataContract]]:
-        """Handle parsing of children in parse in _parse_hierarchical
-        Outsourcing logic for the sole purpose of R0912: Too many branches (15/12) (too-many-branches)
+        """Handle the parsing of child nodes during hierarchical parsing.
 
-        Args:
-            doc (DocumentNode)
-            child (DocumentNode)
-            text_w_prev_child (str)
-            remaining_snipped (str)
-            recursive_depth (int): internal recursion-depth tracker
+        This method is used to process child nodes of a document node and determine
+        how their text content should be handled based on specific conditions. It
+        helps manage recursion depth and ensures the text is split into manageable
+        Markdown data contracts.
+
+            doc (DocumentNode): The parent document node containing metadata and text.
+            child (DocumentNode): The child document node to be processed.
+            text_w_prev_child (str): The text content combined with the previous child node.
+            remaining_snipped (str): The remaining text snippet to be processed.
+            recursive_depth (int): Tracks the current recursion depth during hierarchical parsing.
+
+            tuple[str, list[MarkdownDataContract]]:
+                - A string representing the remaining unprocessed text snippet.
+                - A list of MarkdownDataContract objects containing processed Markdown data.
 
-        Returns:
-            tuple[str, list[DocumentNode]]: remaining_snipped, MdContracts
         """
         return_doc = []
         if self._is_short(text_w_prev_child):
@@ -481,7 +510,7 @@ class SemanticSplitter:
             child["text"] = text_w_prev_child
             return_doc += [
                 MarkdownDataContract(
-                    md=_cut_to_tokenlen(child["text"], self.token_limit),
+                    md=self._cut_to_tokenlen(child["text"], self.token_limit),
                     url=child["metadata"]["url"],
                     keywords=child["metadata"]["keywords"],
                 )
@@ -489,13 +518,11 @@ class SemanticSplitter:
             remaining_snipped = ""
         else:
             if not _is_standalone_a_heading(remaining_snipped):
-                if _get_token_len(remaining_snipped) >= self.token_limit_min:
+                if self._get_token_len(remaining_snipped) >= self.token_limit_min:
                     return_doc.append(
                         MarkdownDataContract(
                             md=remaining_snipped,
-                            keywords=doc["metadata"][
-                                "keywords"
-                            ],
+                            keywords=doc["metadata"]["keywords"],
                             url=doc["metadata"]["url"],
                         )
                     )
@@ -516,9 +543,16 @@ class SemanticSplitter:
                 ]
         return remaining_snipped, return_doc
 
-    def _md_data_from_dict_cut(self, doc):
+    def _md_data_from_dict_cut(self, doc: DocumentNode) -> MarkdownDataContract:
+        """Cut text of Markdown document to fit token limit."""
+        text, discarded_text = self._cut_to_tokenlen(doc["text"], self.token_limit, return_discarded_text=True)
+
+        if len(discarded_text) > 0:
+            # Log document and discarded text
+            log.warning("Splitter cut off text from source document", extra={"discarded_text": discarded_text, **doc})
+
         return MarkdownDataContract(
-            md=_cut_to_tokenlen(doc["text"], self.token_limit),
+            md=text,
             url=doc["metadata"]["url"],
             keywords=doc["metadata"]["keywords"],
         )
@@ -528,7 +562,7 @@ class SemanticSplitter:
         doc: DocumentNode,
         recursive_depth: int = 1,
     ) -> list[MarkdownDataContract]:
-        if _get_token_len(doc["text"]) <= self.token_limit_min:
+        if self._get_token_len(doc["text"]) <= self.token_limit_min:
             if recursive_depth == 1:
                 return [self._md_data_from_dict_cut(doc)]
             log.warning("document to short", extra=doc)
@@ -542,11 +576,20 @@ class SemanticSplitter:
             )
             return [self._md_data_from_dict_cut(doc)]
         if self._is_table(doc):
-            log.warning(
-                "found table, that should have been split, cutting off",
-                extra=doc["metadata"],
+            # split table into chunks depending on token length (re-emit table header)
+            table_splitter = MarkdownTableSplitterUtil(
+                token_limit=self.token_limit,
+                enc=self.tokenizer_model_encoding,
+                repeat_header_row=self.repeat_table_header_row,
             )
-            return [self._md_data_from_dict_cut(doc)]
+            return [
+                MarkdownDataContract(
+                    md=chunk_md,
+                    url=doc["metadata"]["url"],
+                    keywords=doc["metadata"]["keywords"],
+                )
+                for chunk_md in table_splitter.split(doc["text"])
+            ]
         if len(doc["children"]) == 0:
             log.warning(
                 "no remaining children. still to big -> Cut by tokenlen",
@@ -565,10 +608,8 @@ class SemanticSplitter:
             highest_child_is_heading = _get_highest_level_of_children(children) <= 6
             first_child_is_highest = idx_highest == 0
             if not first_child_is_highest and highest_child_is_heading:
-                text_until_highest_child = "\n".join(
-                    [c["text"] for c in children[:idx_highest]]
-                )
-                token_len = _get_token_len(text_until_highest_child)
+                text_until_highest_child = "\n".join([c["text"] for c in children[:idx_highest]])
+                token_len = self._get_token_len(text_until_highest_child)
                 max_level = max(c["highest_level"] for c in children[:idx_highest])
                 new_doc = DocumentNode(
                     highest_level=max_level,
@@ -590,10 +631,10 @@ class SemanticSplitter:
             return_doc += returned_docs
 
         # add potential short remaining spillovers
-        if _get_token_len(remaining_snipped) >= self.token_limit_min:
+        if self._get_token_len(remaining_snipped) >= self.token_limit_min:
             return_doc += [
                 MarkdownDataContract(
-                    md=_cut_to_tokenlen(remaining_snipped, self.token_limit),
+                    md=self._cut_to_tokenlen(remaining_snipped, self.token_limit),
                     url=doc["metadata"]["url"],
                     keywords=doc["metadata"]["keywords"],
                 )
@@ -604,8 +645,7 @@ class SemanticSplitter:
         self,
         docs: list[MarkdownDataContract],
     ) -> list[MarkdownDataContract]:
-        """
-        Function to improve the semantic meaning of the Markdown document by reattaching a parent heading
+        """Function to improve the semantic meaning of the Markdown document by reattaching a parent heading.
 
         Does not yet respect the token limit, however headings usually have little impact
         """
@@ -614,85 +654,34 @@ class SemanticSplitter:
             text = doc.md.strip()
             md_doc = MisDocument(text)
             # Create Header
-            highest_level, highest_type, highest_child = self._find_highest_level(
-                md_doc.children
-            )
+            highest_level, highest_type, highest_child = self._find_highest_level(md_doc.children)
             if str(highest_type) == str(block_token.Heading):
                 # Discuss this
-                #highest_header_until_now[highest_level] = _get_heading_text(highest_child)
+                # highest_header_until_now[highest_level] = _get_heading_text(highest_child)
                 highest_header_until_now[highest_level] = self._render_doc(highest_child).lstrip(" #")
 
-            ordered_headers = {
-                level: text for level, text in highest_header_until_now.items() if text
-            }
-            docwide_highest_level = (
-                min(ordered_headers.keys()) if ordered_headers else 10
-            )
-            new_header = " - ".join(
-                [
-                    text
-                    for level, text in ordered_headers.items()
-                    if level < highest_level
-                ]
-            )
+            ordered_headers = {level: text for level, text in highest_header_until_now.items() if text}
+            docwide_highest_level = min(ordered_headers.keys()) if ordered_headers else 10
+            new_header = " - ".join([text for level, text in ordered_headers.items() if level < highest_level])
             if new_header:
                 new_header = "# " + new_header
             # Filter Doc
-            document_is_just_single_header = (
-                text.strip().startswith("#") and "\n" not in text.strip()
-            )
+            document_is_just_single_header = text.strip().startswith("#") and "\n" not in text.strip()
             if document_is_just_single_header:
                 continue
             # TODOo check token limit or limit header lenght
             # The higher heading the lower its level
             doc_has_lower_heading = docwide_highest_level < highest_level
-            new_doc = (
-                text.strip()
-                if not doc_has_lower_heading
-                else (new_header + "\n\n" + doc.md).strip()
-            )
+            new_doc = text.strip() if not doc_has_lower_heading else (new_header + "\n\n" + doc.md).strip()
             doc.md = new_doc
 
         return docs
 
-    def _remove_irrelevant_nodes(self, doc: MisDocument) -> MisDocument:
-        """Remove irrelevant nodes for LLMs from Mistletoe document"""
-        if not hasattr(doc, "children"):
-            return doc
-        if doc.children is None:
-            return doc
-        cleaned_children = [
-            self._remove_irrelevant_nodes(x)
-            for x in doc.children
-            if not isinstance(x, (span_token.Image, block_token.ThematicBreak))
-        ]
-        doc.children = cleaned_children
-        return doc
-
-    def _remove_irrelevant_nodes_from_str(self, text: str) -> str:
-        """Remove irrelevant nodes for LLMs from string"""
-        mistle_doc = MisDocument(text.strip())
-        new_doc = self._remove_irrelevant_nodes(mistle_doc)
-        with MarkdownRenderer() as renderer:
-            return renderer.render(new_doc).strip().strip(" -\n")  # type: ignore[no-any-return]
-
-    def split_markdown_document(
-        self, doc: MarkdownDataContract
-    ) -> list[MarkdownDataContract]:
-        """Split a Markdown Document into Snippets"""
-        # Disabling this for now due to https://github.com/miyuchina/mistletoe/issues/211
-        _ = self._remove_irrelevant_nodes_from_str(doc.md)
-        metadata = MetaDataDict(
-            url=doc.url, keywords=doc.keywords
-        )
+    def split_markdown_document(self, doc: MarkdownDataContract) -> list[MarkdownDataContract]:
+        """Split a Markdown Document into Snippets."""
+        metadata = MetaDataDict(url=doc.url, keywords=doc.keywords)
         doc_hierarchy: DocumentNode = self._markdown_hierarchy_parser(doc.md, metadata)
-        doc_snippets: list[MarkdownDataContract] = self._parse_hierarchical(
-            doc_hierarchy
-        )
-        improved_snippets: list[MarkdownDataContract] = (
-            self._adopt_splitted_list_to_use_highest_prev_header(doc_snippets)
-        )
-        formatted_snippets: list[MarkdownDataContract] = _format_markdown_docs(
-            improved_snippets
-        )
+        doc_snippets: list[MarkdownDataContract] = self._parse_hierarchical(doc_hierarchy)
+        improved_snippets: list[MarkdownDataContract] = self._adopt_splitted_list_to_use_highest_prev_header(doc_snippets)
+        formatted_snippets: list[MarkdownDataContract] = _format_markdown_docs(improved_snippets)
         return formatted_snippets
